@@ -1,4 +1,4 @@
-// Copyright (c) 2016-2024 The Regents of the University of Michigan
+// Copyright (c) 2016-2025 The Regents of the University of Michigan
 // Part of GSD, released under the BSD 2-Clause License.
 
 #include <sys/stat.h>
@@ -65,16 +65,10 @@ enum
     GSD_INITIAL_WRITE_BUFFER_SIZE = 1024
     };
 
-/// Default maximum size of write buffer
+/// Default maximum size of write buffer (in bytes)
 enum
     {
-    GSD_DEFAULT_MAXIMUM_WRITE_BUFFER_SIZE = 64 * 1024 * 1024
-    };
-
-/// Default number of index entries to buffer
-enum
-    {
-    GSD_DEFAULT_INDEX_ENTRIES_TO_BUFFER = 256 * 1024
+    GSD_DEFAULT_MAXIMUM_WRITE_BUFFER_BYTES = 1024 * 1024
     };
 
 /// Size of hash map
@@ -98,7 +92,6 @@ enum
 #ifdef _WIN32
 #define lseek _lseeki64
 #define ftruncate _chsize
-#define fsync _commit
 typedef int64_t ssize_t;
 
 int S_IRUSR = _S_IREAD;
@@ -106,6 +99,11 @@ int S_IWUSR = _S_IWRITE;
 int S_IRGRP = _S_IREAD;
 int S_IWGRP = _S_IWRITE;
 
+int gsd_fsync(int fd)
+    {
+    return _commit(fd);
+    }
+
 inline ssize_t pread(int fd, void* buf, size_t count, int64_t offset)
     {
     // Note: _read only accepts unsigned int values
@@ -132,6 +130,16 @@ inline ssize_t pwrite(int fd, const void* buf, size_t count, int64_t offset)
     return result;
     }
 
+#elif defined(__APPLE__)
+int gsd_fsync(int fd)
+    {
+    return fcntl(fd, F_FULLFSYNC);
+    }
+#else
+int gsd_fsync(int fd)
+    {
+    return fsync(fd);
+    }
 #endif
 
 /** Zero memory
@@ -177,6 +185,7 @@ inline static ssize_t gsd_io_pwrite_retry(int fd, const void* buf, size_t count,
         errno = 0;
         ssize_t bytes_written
             = pwrite(fd, ptr + total_bytes_written, to_write, offset + total_bytes_written);
+
         if (bytes_written == -1 || (bytes_written == 0 && errno != 0))
             {
             return GSD_ERROR_IO;
@@ -624,25 +633,23 @@ inline static int gsd_index_buffer_map(struct gsd_index_buffer* buf, struct gsd_
 
 #if GSD_USE_MMAP
     // map the index in read only mode
-    size_t page_size = getpagesize();
     size_t index_size = sizeof(struct gsd_index_entry) * handle->header.index_allocated_entries;
-    size_t offset = (handle->header.index_location / page_size) * page_size;
     buf->mapped_data = mmap(NULL,
-                            index_size + (handle->header.index_location - offset),
+                            index_size + handle->header.index_location,
                             PROT_READ,
                             MAP_SHARED,
                             handle->fd,
-                            offset);
+                            0);
 
     if (buf->mapped_data == MAP_FAILED)
         {
         return GSD_ERROR_IO;
         }
 
-    buf->data = (struct gsd_index_entry*)(((char*)buf->mapped_data)
-                                          + (handle->header.index_location - offset));
+    buf->data
+        = (struct gsd_index_entry*)(((char*)buf->mapped_data) + handle->header.index_location);
 
-    buf->mapped_len = index_size + (handle->header.index_location - offset);
+    buf->mapped_len = index_size + handle->header.index_location;
     buf->reserved = handle->header.index_allocated_entries;
 #else
     // mmap not supported, read the data from the disk
@@ -976,18 +983,25 @@ inline static int gsd_expand_file_index(struct gsd_handle* handle, size_t size_r
         return retval;
         }
 
-    // allocate the copy buffer
-    uint64_t copy_buffer_size
-        = GSD_DEFAULT_INDEX_ENTRIES_TO_BUFFER * sizeof(struct gsd_index_entry);
+    int64_t new_index_location = lseek(handle->fd, 0, SEEK_END);
+    int64_t old_index_location = handle->header.index_location;
+
+    // fill the new index space with 0s
+    size_t new_index_bytes = size_new * sizeof(struct gsd_index_entry);
+    retval = ftruncate(handle->fd, new_index_location + new_index_bytes);
+    if (retval != 0)
+        {
+        return GSD_ERROR_IO;
+        }
+
+    // write the current index to the end of the file
+    uint64_t copy_buffer_size = GSD_DEFAULT_MAXIMUM_WRITE_BUFFER_BYTES;
     if (copy_buffer_size > size_old * sizeof(struct gsd_index_entry))
         {
         copy_buffer_size = size_old * sizeof(struct gsd_index_entry);
         }
     char* buf = malloc(copy_buffer_size);
 
-    // write the current index to the end of the file
-    int64_t new_index_location = lseek(handle->fd, 0, SEEK_END);
-    int64_t old_index_location = handle->header.index_location;
     size_t total_bytes_written = 0;
     size_t old_index_bytes = size_old * sizeof(struct gsd_index_entry);
     while (total_bytes_written < old_index_bytes)
@@ -1023,34 +1037,8 @@ inline static int gsd_expand_file_index(struct gsd_handle* handle, size_t size_r
         total_bytes_written += bytes_written;
         }
 
-    // fill the new index space with 0s
-    gsd_util_zero_memory(buf, copy_buffer_size);
-
-    size_t new_index_bytes = size_new * sizeof(struct gsd_index_entry);
-    while (total_bytes_written < new_index_bytes)
-        {
-        size_t bytes_to_copy = copy_buffer_size;
-        if (new_index_bytes - total_bytes_written < copy_buffer_size)
-            {
-            bytes_to_copy = new_index_bytes - total_bytes_written;
-            }
-
-        ssize_t bytes_written = gsd_io_pwrite_retry(handle->fd,
-                                                    buf,
-                                                    bytes_to_copy,
-                                                    new_index_location + total_bytes_written);
-
-        if (bytes_written == -1 || bytes_written != bytes_to_copy)
-            {
-            free(buf);
-            return GSD_ERROR_IO;
-            }
-
-        total_bytes_written += bytes_written;
-        }
-
     // sync the expanded index
-    retval = fsync(handle->fd);
+    retval = gsd_fsync(handle->fd);
     if (retval != 0)
         {
         free(buf);
@@ -1062,7 +1050,7 @@ inline static int gsd_expand_file_index(struct gsd_handle* handle, size_t size_r
 
     // update the header
     handle->header.index_location = new_index_location;
-    handle->file_size = handle->header.index_location + total_bytes_written;
+    handle->file_size = new_index_location + new_index_bytes;
     handle->header.index_allocated_entries = size_new;
 
     // write the new header out
@@ -1074,7 +1062,7 @@ inline static int gsd_expand_file_index(struct gsd_handle* handle, size_t size_r
         }
 
     // sync the updated header
-    retval = fsync(handle->fd);
+    retval = gsd_fsync(handle->fd);
     if (retval != 0)
         {
         return GSD_ERROR_IO;
@@ -1094,10 +1082,9 @@ inline static int gsd_expand_file_index(struct gsd_handle* handle, size_t size_r
     @brief Flush the write buffer.
 
     gsd_write_frame() writes small data chunks into the write buffer. It adds index entries for
-    these chunks to gsd_handle::buffer_index with locations offset from the start of the write
-    buffer. gsd_flush_write_buffer() writes the buffer to the end of the file, moves the index
-    entries to gsd_handle::frame_index and updates the location to reference the beginning of the
-    file.
+    these chunks to gsd_handle::buffer_index with locations offset from the current end of file.
+    gsd_flush_write_buffer() writes the buffer to the end of the file. The entries in the buffered
+    index have valid locations because nothing else writes past the end of the file.
 
     @param handle Handle to flush the write buffer.
     @returns GSD_SUCCESS on success or GSD_* error codes on error
@@ -1109,18 +1096,12 @@ inline static int gsd_flush_write_buffer(struct gsd_handle* handle)
         return GSD_ERROR_INVALID_ARGUMENT;
         }
 
-    if (handle->write_buffer.size == 0 && handle->buffer_index.size == 0)
+    if (handle->write_buffer.size == 0)
         {
         // nothing to do
         return GSD_SUCCESS;
         }
 
-    if (handle->write_buffer.size > 0 && handle->buffer_index.size == 0)
-        {
-        // error: bytes in buffer, but no index for them
-        return GSD_ERROR_INVALID_ARGUMENT;
-        }
-
     // write the buffer to the end of the file
     uint64_t offset = handle->file_size;
     ssize_t bytes_written = gsd_io_pwrite_retry(handle->fd,
@@ -1138,24 +1119,6 @@ inline static int gsd_flush_write_buffer(struct gsd_handle* handle)
     // reset write_buffer for new data
     handle->write_buffer.size = 0;
 
-    // Move buffer_index entries to frame_index.
-    size_t i;
-    for (i = 0; i < handle->buffer_index.size; i++)
-        {
-        struct gsd_index_entry* new_index;
-        int retval = gsd_index_buffer_add(&handle->frame_index, &new_index);
-        if (retval != GSD_SUCCESS)
-            {
-            return retval;
-            }
-
-        *new_index = handle->buffer_index.data[i];
-        new_index->location += offset;
-        }
-
-    // clear the buffer index for new entries
-    handle->buffer_index.size = 0;
-
     return GSD_SUCCESS;
     }
 
@@ -1226,7 +1189,7 @@ inline static int gsd_flush_name_buffer(struct gsd_handle* handle)
             }
 
         // sync the updated name list
-        retval = fsync(handle->fd);
+        retval = gsd_fsync(handle->fd);
         if (retval != 0)
             {
             return GSD_ERROR_IO;
@@ -1260,7 +1223,7 @@ inline static int gsd_flush_name_buffer(struct gsd_handle* handle)
         }
 
     // sync the updated name list or header
-    retval = fsync(handle->fd);
+    retval = gsd_fsync(handle->fd);
     if (retval != 0)
         {
         return GSD_ERROR_IO;
@@ -1433,7 +1396,7 @@ gsd_initialize_file(int fd, const char* application, const char* schema, uint32_
         }
 
     // sync file
-    retval = fsync(fd);
+    retval = gsd_fsync(fd);
     if (retval != 0)
         {
         return GSD_ERROR_IO;
@@ -1573,22 +1536,17 @@ inline static int gsd_initialize_handle(struct gsd_handle* handle)
     // determine the current frame counter
     if (handle->file_index.size == 0)
         {
-        handle->cur_frame = 0;
+        handle->file_frame = 0;
         }
     else
         {
-        handle->cur_frame = handle->file_index.data[handle->file_index.size - 1].frame + 1;
+        handle->file_frame = handle->file_index.data[handle->file_index.size - 1].frame + 1;
         }
+    handle->buffer_frame = handle->file_frame;
 
     // if this is a write mode, allocate the initial frame index and the name buffer
     if (handle->open_flags != GSD_OPEN_READONLY)
         {
-        retval = gsd_index_buffer_allocate(&handle->frame_index, GSD_INITIAL_FRAME_INDEX_SIZE);
-        if (retval != GSD_SUCCESS)
-            {
-            return retval;
-            }
-
         retval = gsd_index_buffer_allocate(&handle->buffer_index, GSD_INITIAL_FRAME_INDEX_SIZE);
         if (retval != GSD_SUCCESS)
             {
@@ -1610,8 +1568,7 @@ inline static int gsd_initialize_handle(struct gsd_handle* handle)
         }
 
     handle->pending_index_entries = 0;
-    handle->maximum_write_buffer_size = GSD_DEFAULT_MAXIMUM_WRITE_BUFFER_SIZE;
-    handle->index_entries_to_buffer = GSD_DEFAULT_INDEX_ENTRIES_TO_BUFFER;
+    handle->maximum_write_buffer_size = GSD_DEFAULT_MAXIMUM_WRITE_BUFFER_BYTES;
 
     // Silently upgrade writable files from a previous matching major version to the latest
     // minor version.
@@ -1804,15 +1761,6 @@ int gsd_truncate(struct gsd_handle* handle)
         return retval;
         }
 
-    if (handle->frame_index.reserved > 0)
-        {
-        retval = gsd_index_buffer_free(&handle->frame_index);
-        if (retval != GSD_SUCCESS)
-            {
-            return retval;
-            }
-        }
-
     if (handle->buffer_index.reserved > 0)
         {
         retval = gsd_index_buffer_free(&handle->buffer_index);
@@ -1873,15 +1821,6 @@ int gsd_close(struct gsd_handle* handle)
         return retval;
         }
 
-    if (handle->frame_index.reserved > 0)
-        {
-        retval = gsd_index_buffer_free(&handle->frame_index);
-        if (retval != GSD_SUCCESS)
-            {
-            return retval;
-            }
-        }
-
     if (handle->buffer_index.reserved > 0)
         {
         retval = gsd_index_buffer_free(&handle->buffer_index);
@@ -1947,14 +1886,9 @@ int gsd_end_frame(struct gsd_handle* handle)
         return GSD_ERROR_FILE_MUST_BE_WRITABLE;
         }
 
-    handle->cur_frame++;
+    handle->buffer_frame++;
     handle->pending_index_entries = 0;
 
-    if (handle->frame_index.size > 0 || handle->buffer_index.size > handle->index_entries_to_buffer)
-        {
-        return gsd_flush(handle);
-        }
-
     return GSD_SUCCESS;
     }
 
@@ -1969,34 +1903,34 @@ int gsd_flush(struct gsd_handle* handle)
         return GSD_ERROR_FILE_MUST_BE_WRITABLE;
         }
 
-    // flush the namelist buffer
-    int retval = gsd_flush_name_buffer(handle);
+    // flush the write buffer first so that all locations in the buffer_index are maintained.
+    int retval = gsd_flush_write_buffer(handle);
     if (retval != GSD_SUCCESS)
         {
         return retval;
         }
 
-    // flush the write buffer
-    retval = gsd_flush_write_buffer(handle);
+    // flush the namelist buffer
+    retval = gsd_flush_name_buffer(handle);
     if (retval != GSD_SUCCESS)
         {
         return retval;
         }
 
     // sync the data before writing the index
-    retval = fsync(handle->fd);
+    retval = gsd_fsync(handle->fd);
     if (retval != 0)
         {
         return GSD_ERROR_IO;
         }
 
-    // Write the frame index to the file, excluding the index entries that are part of the current
+    // Write the buffer index to the file, excluding the index entries that are part of the current
     // frame.
-    if (handle->pending_index_entries > handle->frame_index.size)
+    if (handle->pending_index_entries > handle->buffer_index.size)
         {
         return GSD_ERROR_INVALID_ARGUMENT;
         }
-    uint64_t index_entries_to_write = handle->frame_index.size - handle->pending_index_entries;
+    uint64_t index_entries_to_write = handle->buffer_index.size - handle->pending_index_entries;
 
     if (index_entries_to_write > 0)
         {
@@ -2007,7 +1941,7 @@ int gsd_flush(struct gsd_handle* handle)
             }
 
         // sort the index before writing
-        retval = gsd_index_buffer_sort(&handle->frame_index);
+        retval = gsd_index_buffer_sort(&handle->buffer_index);
         if (retval != 0)
             {
             return retval;
@@ -2019,17 +1953,23 @@ int gsd_flush(struct gsd_handle* handle)
 
         size_t bytes_to_write = sizeof(struct gsd_index_entry) * index_entries_to_write;
         ssize_t bytes_written
-            = gsd_io_pwrite_retry(handle->fd, handle->frame_index.data, bytes_to_write, write_pos);
+            = gsd_io_pwrite_retry(handle->fd, handle->buffer_index.data, bytes_to_write, write_pos);
 
         if (bytes_written == -1 || bytes_written != bytes_to_write)
             {
             return GSD_ERROR_IO;
             }
 
+        retval = gsd_fsync(handle->fd);
+        if (retval != 0)
+            {
+            return GSD_ERROR_IO;
+            }
+
 #if !GSD_USE_MMAP
         // add the entries to the file index
         memcpy(handle->file_index.data + handle->file_index.size,
-               handle->frame_index.data,
+               handle->buffer_index.data,
                sizeof(struct gsd_index_entry) * index_entries_to_write);
 #endif
 
@@ -2041,15 +1981,17 @@ int gsd_flush(struct gsd_handle* handle)
             {
             for (uint64_t i = 0; i < handle->pending_index_entries; i++)
                 {
-                handle->frame_index.data[i]
-                    = handle->frame_index
-                          .data[handle->frame_index.size - handle->pending_index_entries + i];
+                handle->buffer_index.data[i]
+                    = handle->buffer_index
+                          .data[handle->buffer_index.size - handle->pending_index_entries + i];
                 }
             }
 
-        handle->frame_index.size = handle->pending_index_entries;
+        handle->buffer_index.size = handle->pending_index_entries;
         }
 
+    handle->file_frame = handle->buffer_frame;
+
     return GSD_SUCCESS;
     }
 
@@ -2099,34 +2041,36 @@ int gsd_write_chunk(struct gsd_handle* handle,
     struct gsd_index_entry entry;
     // populate fields in the entry's data
     gsd_util_zero_memory(&entry, sizeof(struct gsd_index_entry));
-    entry.frame = handle->cur_frame;
+    entry.frame = handle->buffer_frame;
     entry.id = id;
     entry.type = (uint8_t)type;
     entry.N = N;
     entry.M = M;
+    entry.location = handle->file_size + handle->write_buffer.size;
     size_t size = N * M * gsd_sizeof_type(type);
 
-    // decide whether to write this chunk to the buffer or straight to disk
-    if (size < handle->maximum_write_buffer_size)
+    // flush the buffer if this entry won't fit
+    if (size > (handle->maximum_write_buffer_size - handle->write_buffer.size))
         {
-        // flush the buffer if this entry won't fit
-        if (size > (handle->maximum_write_buffer_size - handle->write_buffer.size))
-            {
-            gsd_flush_write_buffer(handle);
-            }
-
-        entry.location = handle->write_buffer.size;
-
-        // add an entry to the buffer index
-        struct gsd_index_entry* index_entry;
-
-        int retval = gsd_index_buffer_add(&handle->buffer_index, &index_entry);
+        int retval = gsd_flush_write_buffer(handle);
         if (retval != GSD_SUCCESS)
             {
             return retval;
             }
-        *index_entry = entry;
+        }
 
+    // add an entry to the buffer index
+    struct gsd_index_entry* index_entry;
+    int retval = gsd_index_buffer_add(&handle->buffer_index, &index_entry);
+    if (retval != GSD_SUCCESS)
+        {
+        return retval;
+        }
+    *index_entry = entry;
+
+    // decide whether to write this chunk to the buffer or straight to disk
+    if (size < handle->maximum_write_buffer_size)
+        {
         // add the data to the write buffer
         if (size > 0)
             {
@@ -2139,21 +2083,8 @@ int gsd_write_chunk(struct gsd_handle* handle,
         }
     else
         {
-        // add an entry to the frame index
-        struct gsd_index_entry* index_entry;
-
-        int retval = gsd_index_buffer_add(&handle->frame_index, &index_entry);
-        if (retval != GSD_SUCCESS)
-            {
-            return retval;
-            }
-        *index_entry = entry;
-
-        // find the location at the end of the file for the chunk
-        index_entry->location = handle->file_size;
-
         // write the data
-        ssize_t bytes_written = gsd_io_pwrite_retry(handle->fd, data, size, index_entry->location);
+        ssize_t bytes_written = gsd_io_pwrite_retry(handle->fd, data, size, entry.location);
         if (bytes_written == -1 || bytes_written != size)
             {
             return GSD_ERROR_IO;
@@ -2173,7 +2104,7 @@ uint64_t gsd_get_nframes(struct gsd_handle* handle)
         {
         return 0;
         }
-    return handle->cur_frame;
+    return handle->file_frame;
     }
 
 const struct gsd_index_entry*
@@ -2191,14 +2122,6 @@ gsd_find_chunk(struct gsd_handle* handle, uint64_t frame, const char* name)
         {
         return NULL;
         }
-    if (handle->open_flags != GSD_OPEN_READONLY)
-        {
-        int retval = gsd_flush(handle);
-        if (retval != GSD_SUCCESS)
-            {
-            return NULL;
-            }
-        }
 
     // find the id for the given name
     uint16_t match_id = gsd_name_id_map_find(&handle->name_map, name);
@@ -2289,14 +2212,6 @@ int gsd_read_chunk(struct gsd_handle* handle, void* data, const struct gsd_index
         {
         return GSD_ERROR_INVALID_ARGUMENT;
         }
-    if (handle->open_flags != GSD_OPEN_READONLY)
-        {
-        int retval = gsd_flush(handle);
-        if (retval != GSD_SUCCESS)
-            {
-            return retval;
-            }
-        }
 
     size_t size = chunk->N * chunk->M * gsd_sizeof_type((enum gsd_type)chunk->type);
     if (size == 0)
@@ -2392,14 +2307,6 @@ gsd_find_matching_chunk_name(struct gsd_handle* handle, const char* match, const
         {
         return NULL;
         }
-    if (handle->open_flags != GSD_OPEN_READONLY)
-        {
-        int retval = gsd_flush(handle);
-        if (retval != GSD_SUCCESS)
-            {
-            return NULL;
-            }
-        }
 
     // return nothing found if the name buffer is corrupt
     if (handle->file_names.data.data[handle->file_names.data.reserved - 1] != 0)
@@ -2468,7 +2375,7 @@ int gsd_upgrade(struct gsd_handle* handle)
         {
         return GSD_ERROR_INVALID_ARGUMENT;
         }
-    if (handle->frame_index.size > 0 || handle->frame_names.n_names > 0)
+    if (handle->buffer_index.size > 0 || handle->frame_names.n_names > 0)
         {
         return GSD_ERROR_INVALID_ARGUMENT;
         }
@@ -2516,7 +2423,7 @@ int gsd_upgrade(struct gsd_handle* handle)
                 }
 
             // sync the updated index
-            retval = fsync(handle->fd);
+            retval = gsd_fsync(handle->fd);
             if (retval != 0)
                 {
                 return GSD_ERROR_IO;
@@ -2574,7 +2481,7 @@ int gsd_upgrade(struct gsd_handle* handle)
             handle->file_names.data = new_name_buf;
 
             // sync the updated name list
-            retval = fsync(handle->fd);
+            retval = gsd_fsync(handle->fd);
             if (retval != 0)
                 {
                 gsd_byte_buffer_free(&new_name_buf);
@@ -2595,7 +2502,7 @@ int gsd_upgrade(struct gsd_handle* handle)
             }
 
         // sync the updated header
-        int retval = fsync(handle->fd);
+        int retval = gsd_fsync(handle->fd);
         if (retval != 0)
             {
             return GSD_ERROR_IO;
@@ -2639,27 +2546,6 @@ int gsd_set_maximum_write_buffer_size(struct gsd_handle* handle, uint64_t size)
     return GSD_SUCCESS;
     }
 
-uint64_t gsd_get_index_entries_to_buffer(struct gsd_handle* handle)
-    {
-    if (handle == NULL)
-        {
-        return 0;
-        }
-    return handle->index_entries_to_buffer;
-    }
-
-int gsd_set_index_entries_to_buffer(struct gsd_handle* handle, uint64_t number)
-    {
-    if (handle == NULL || number == 0)
-        {
-        return GSD_ERROR_INVALID_ARGUMENT;
-        }
-
-    handle->index_entries_to_buffer = number;
-
-    return GSD_SUCCESS;
-    }
-
 // undefine windows wrapper macros
 #ifdef _WIN32
 #undef lseek
